import json
import logging
import traceback
from dataclasses import is_dataclass, asdict
from datetime import date, datetime, time
from decimal import Decimal
from inspect import istraceback
from typing import Dict, Optional, IO, Union, Any, List

LOG_PROTECTED_KWARGS = ("exc_info", "stack_info", "stacklevel", "extra")
LOG_RECORD_FIELDS = (
    "args",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "msecs",
    "msg",
    "pathname",
    "process",
    "processName",
    "stack_info",
    "thread",
    "threadName",
)


class _LogArgsBucket:
    def __init__(self, args: Dict[str, Any]) -> None:
        self._args = args

    def __getitem__(self, item: str) -> Any:
        return self.__getattr__(item)

    def __getattr__(self, item: str) -> Any:
        if item in self._args:
            if isinstance(self._args[item], dict):
                return _LogArgsBucket(self._args[item])

            return str(self._args[item])
        else:
            return ""

    def __str__(self) -> str:
        return str(self._args)


class JsonEncoder(json.JSONEncoder):
    def default(self, data: Any) -> Any:
        if isinstance(data, (date, datetime, time)):
            return data.isoformat()
        elif istraceback(data):
            return "".join(traceback.format_tb(data)).strip()
        elif is_dataclass(data):
            return asdict(data)
        elif isinstance(data, Exception):
            return str(data)

        try:
            return str(data)
        except Exception:
            return repr(data)


class JsonFormatter(logging.Formatter):
    def __init__(
        self, json_encoder: json.JSONEncoder = JsonEncoder(), message_format: str = "[{level}] {timestamp} {msg}"
    ):
        self.json_encoder = json_encoder
        self.message_format = message_format
        super(JsonFormatter, self).__init__()

    @staticmethod
    def get_message(record: logging.LogRecord) -> Any:
        if isinstance(record.msg, (str, float, int, bool, Decimal, type(None))):
            return record.msg

        if record.levelname != "DEBUG":
            return repr(record.msg)

        return record.msg

    @staticmethod
    def format_tags(record: logging.LogRecord) -> Dict[str, str]:
        if hasattr(record, "tags"):
            return {**getattr(record, "tags"), "source_path": f"{record.module}.{record.funcName}:{record.lineno}"}

        return {"path": f"{record.module}.{record.funcName}:{record.lineno}"}

    def format_message(self, log_entry: dict) -> str:
        return self.message_format.format_map(_LogArgsBucket(log_entry))

    def format(self, record: logging.LogRecord) -> str:
        message = self.get_message(record)

        if hasattr(record, "_message_kwargs") and record._message_kwargs:  # type: ignore
            msg = message.format(**record._message_kwargs)  # type: ignore
        else:
            msg = message

        payload = {
            "log_message": message,
            "args": getattr(record, "_message_kwargs", {}),
            "level": record.levelname,
            "timestamp": datetime.utcfromtimestamp(record.created).isoformat(),
            "tags": self.format_tags(record),
        }

        log = {}
        for key in LOG_RECORD_FIELDS:
            if not hasattr(record, key):
                continue
            log[key] = getattr(record, key)

        log = {**log, **payload, **{"msg": msg}}

        return self.format_message(log) + "\t" + json.dumps(payload, cls=JsonEncoder, ensure_ascii=True)


class Logger(logging.Logger):
    _tags: Dict[str, Union[str, List[str], Dict[str, Union[str, List[str], dict]]]] = {}
    _cache: Dict[str, "Logger"] = {}

    @classmethod
    def set_tag(cls, key: str, value: Union[str, List[str], Dict[str, Union[str, List[str], dict]]]) -> None:
        cls._tags[key] = value

    def handle(self, record: logging.LogRecord) -> None:
        setattr(record, "tags", self._tags)
        super(Logger, self).handle(record)

    def _log(self, *args, **kwargs) -> None:
        new_kwargs: Dict[str, Any] = {"extra": {"_message_kwargs": {}}}
        for key, value in kwargs.items():
            if key in LOG_PROTECTED_KWARGS:
                new_kwargs[key] = value
                if key == "extra":
                    new_kwargs["extra"]["_message_kwargs"] = {}
                continue
            new_kwargs["extra"]["_message_kwargs"][key] = value

        super(Logger, self)._log(*args, **new_kwargs)

    @classmethod
    def get(
        cls,
        name: str,
        level: Union[str, int, None] = None,
        log_stream: Optional[IO[str]] = None,
        message_format: str = "[{level}] {timestamp} {msg}",
        propagate: bool = False,
    ) -> "Logger":
        if name in cls._cache:
            return cls._cache[name]

        logger = logging.getLogger(name)

        # Something went wrong with logger's lock mechanism, we should clear handlers to prevent double logs.
        if len(logger.handlers) >= 1:
            logger.handlers.clear()

        json_handler = logging.StreamHandler(log_stream)
        json_handler.setFormatter(JsonFormatter(message_format=message_format))
        logger.addHandler(json_handler)

        logger.setLevel(level or logging.DEBUG)
        logger.propagate = propagate
        cls._cache[name] = logger  # type: ignore

        return logger  # type: ignore


# initialise logging manager with package's class
logging.setLoggerClass(Logger)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['chocs_middleware', 'chocs_middleware.trace']

package_data = \
{'': ['*']}

install_requires = \
['chocs>=1.5.3,<2.0.0', 'gid>=1.0.1,<2.0.0']

setup_kwargs = {
    'name': 'chocs-middleware.trace',
    'version': '0.4.0',
    'description': 'Http tracing middleware for chocs library.',
    'long_description': '# Chocs-Trace <br> [![PyPI version](https://badge.fury.io/py/chocs-middleware.trace.svg)](https://pypi.org/project/chocs-middleware.trace/) [![CI](https://github.com/kodemore/chocs-trace/actions/workflows/main.yaml/badge.svg)](https://github.com/kodemore/chocs-trace/actions/workflows/main.yaml) [![Release](https://github.com/kodemore/chocs-trace/actions/workflows/release.yml/badge.svg)](https://github.com/kodemore/chocs-trace/actions/workflows/release.yml) [![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\nHttp tracing middleware for chocs library. \n\n\n# Installation\n\n### Poetry:\n```bash\npoetry add chocs-middleware.trace\n```\n\n### Pip:\n```bash\npip install chocs-middleware.trace\n```\n\n# Features\n\n- Automatic generation and propagation of tracing headers (x-request-id, x-correlation-id, x-causation-id)\n- Sentry integration\n- More intuitive log formatting options\n- Structured logging\n\n# Usage\n\n## Support tracing in your responses\n\n```python\nfrom chocs_middleware.trace import TraceMiddleware\nfrom chocs import Application, HttpRequest, HttpResponse\n\n# id_prefix will ensure generated tracing headers to contain your prefix\napp = Application(TraceMiddleware(id_prefix="service-name-"))\n\n\n@app.get("/hello")\ndef say_hello(req: HttpRequest) -> HttpResponse:\n    return HttpResponse("Hello!")  # tracing middleware will automatically attach x-request-id, x-correlation-id, x-causation-id headers to your response\n\n```\n\n## Tracing requests\n\n```python\nfrom chocs_middleware.trace import TraceMiddleware, HttpStrategy\nfrom chocs import Application, HttpRequest, HttpResponse\nimport requests\n\n# http_strategy will try to detect requests library and use it to add tracing headers in all your requests\n# if it fails to detect requests library it will fallback to urllib3\napp = Application(TraceMiddleware(http_strategy=HttpStrategy.AUTO))\n\n\n@app.get("/hello")\ndef say_hello(req: HttpRequest) -> HttpResponse:\n    \n    requests.get("http://example.com/test")  # middleware will automatically attach x-correlation-id, x-causation-id and x-request-id headers to your request\n    \n    return HttpResponse("Hello!")\n```\n\n## Using logger\n\n```python\nfrom chocs import Application, HttpRequest, HttpResponse\nfrom chocs_middleware.trace import TraceMiddleware, Logger\n\napp = Application(TraceMiddleware())\n\n\n@app.get("/hello")\ndef say_hello(req: HttpRequest) -> HttpResponse:\n    logger = Logger.get("logger_name")\n    logger.info("Hello {name}!", name="Bob")  # will output to log stream Hello Bob!\n    return HttpResponse("Hello!")\n```\n\n### Formatting message\n\n```python\nfrom chocs import Application, HttpRequest, HttpResponse\nfrom chocs_middleware.trace import TraceMiddleware, Logger\n\napp = Application(TraceMiddleware())\n\n\n@app.get("/hello")\ndef say_hello(req: HttpRequest) -> HttpResponse:\n    logger = Logger.get("logger_name", message_format="[{level}] {tags.request.x-correlation-id} {msg}")\n    logger.info("Hello {name}!", name="Bob")  # will output to log stream Hello Bob!\n    return HttpResponse("Hello!")\n```\n\n#### Available formatting options\n\n| Name | Example value | Description |\n|---|:---:|---|\n| `{level}` | DEBUG | Log level name |\n| `{msg}` | Example message | Log message after interpolation |\n| `{log_message}` | Example {name} | Log message before interpolation |\n| `{timestamp}` | 2022-03-07T20:06:23.453866 | Time of the logged message |\n| `{filename}` | example.py | Name of the python file where message was log |\n| `{funcName}` | example_function | Name of the function where message was log |\n| `{module}` | example_module | Name of the module where message was log |\n| `{pathname}` | example/path | Path name of the file where message was log |\n| `{tags.*}` | some value | Custom tag value set by calling `Logger.set_tag` function |\n\n\n',
    'author': 'Dawid Kraczkowski',
    'author_email': 'dawid.kraczkowski@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kodemore/chocs-trace',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
